/*
 * Copyright (C) 2013 Noralf Tronnes
 * Copyright (C) 2016 Broadcom
 *
 * SPDX-License-Identifier:      GPL-2.0+
 *
 */

#ifndef __FBTFT_CORE_H
#define __FBTFT_CORE_H

#include <spi.h>

#define FBTFT_GAMMA_MAX_VALUES_TOTAL 128

struct fbtft_par;

/**
 * struct fbtft_ops - FBTFT operations structure
 * @write: Writes to interface bus
 * @write_vmem: Writes video memory to display
 * @write_reg: Writes to controller register
 * @set_addr_win: Set the GRAM update window
 * @reset: Reset the LCD controller
 * @update_display: Updates the display
 * @init_display: Initializes the display
 * @set_var: Configure LCD with values from variables like @rotate and @bgr
 *           (optional)
 * @set_gamma: Set Gamma curve (optional)
 */
struct fbtft_ops {
	int (*write)(struct fbtft_par *par, void *buf, size_t len);
	int (*write_vmem)(struct fbtft_par *par, size_t offset, size_t len);
	void (*write_register)(struct fbtft_par *par, int len, ...);

	void (*set_addr_win)(struct fbtft_par *par,
		int xs, int ys, int xe, int ye);
	void (*reset)(struct fbtft_par *par);
	void (*update_display)(struct fbtft_par *par);
	int (*init_display)(struct fbtft_par *par);

	int (*set_var)(struct fbtft_par *par);
	int (*set_gamma)(struct fbtft_par *par);
};

/**
 * struct fbtft_display - Describes the display properties
 * @width: Width of display in pixels
 * @height: Height of display in pixels
 * @regwidth: LCD Controller Register width in bits
 * @buswidth: Display interface bus width in bits
 * @fbtftops: FBTFT operations provided by driver or device (platform_data)
 * @bpp: Bits per pixel
 * @gamma.curves: Gamma curve(s) values
 * @gamma.num_curves: Number of Gamma curves
 * @gamma.num_values: Number of values per Gamma curve
 */
struct fbtft_display {
	unsigned width;
	unsigned height;
	unsigned regwidth;
	unsigned buswidth;
	struct fbtft_ops fbtftops;
	unsigned bpp;
	struct {
		unsigned long curves[FBTFT_GAMMA_MAX_VALUES_TOTAL];
		int num_curves;
		int num_values;
	} gamma;
};

/**
 * struct fbtft_par - Main FBTFT data structure
 * @spi_bus: SPI bus number
 * @spi_cs: SPI chip select
 * @spi_max_hz: SPI maximum frequency
 * @slave: SPI slave device
 * @display: Dislpay properties
 * @buf: Small buffer used when writing init data over SPI
 * @startbyte: Used by some controllers when in SPI mode.
 *             Format: 6 bit Device id + RS bit + RW bit
 * @gpio.reset: GPIO used to reset display
 * @gpio.dc: Data/Command signal, also known as RS
 * @bgr: BGR mode
 * @rotate: Display rotation angle
 * @fb: frame buffer
 */
struct fbtft_par {
	int spi_bus;
	int spi_cs;
	int spi_max_hz;
	struct spi_slave *slave;
	struct fbtft_display *display;
	u8 buf[128];
	u8 startbyte;
	struct {
		int reset;
		int dc;
	} gpio;
	bool bgr;
	unsigned rotate;
	void *fb;
};

#define NUMARGS(...)  (sizeof((int[]){__VA_ARGS__})/sizeof(int))

#define write_reg(par, ...) \
	par->display->fbtftops.write_register(par, NUMARGS(__VA_ARGS__), \
					      __VA_ARGS__)

#define BIT(x) (1 << (x))

#ifdef FBTFT_DEBUG
#define fbtft_dbg(format, arg...) printf(format, ##arg)
#define fbtft_dbg_hex(type, buf, num, format, arg...)  \
	do {                                           \
		int i;                                 \
		type *ptr = (type *)buf;               \
		printf(format, ##arg);                 \
		for (i = 0; i < num; i++) {            \
			if (i % 16 == 0)               \
				printf("\n");          \
			printf("%x ", ptr[i]);         \
		}                                      \
		printf("\n");                          \
	} while (0)
#else
#define fbtft_dbg(format, arg...)
#define fbtft_dbg_hex(type, buf, num, format, arg...)
#endif /* CONFIG_FBTFT_DEBUG */

#endif /* __FBTFT_CORE_H */
