/*
 * Copyright (C) 2013 Noralf Tronnes
 * Copyright (C) 2013 Christian Vogelgsang
 * Copyright (C) 2016 Broadcom
 *
 * SPDX-License-Identifier:      GPL-2.0+
 *
 */

/*
 * FB driver for the ILI9341 LCD display controller
 */

#include <common.h>
#include <lcd.h>
#include "mipi_display.h"
#include "fbtft_core.h"
#include "fbtft_config.h"

#define WIDTH		240
#define HEIGHT		320
#define DEFAULT_GAMMA	{0x1F, 0x1A, 0x18, 0x0A, 0x0F, 0x06, 0x45, 0x87, \
			 0x32, 0x0A, 0x07, 0x02, 0x07, 0x05, 0x00, \
			 0x00, 0x25, 0x27, 0x05, 0x10, 0x09, 0x3A, 0x78, \
			 0x4D, 0x05, 0x18, 0x0D, 0x38, 0x3A, 0x1F }

static int init_display(struct fbtft_par *par)
{
	fbtft_dbg("%s()\n", __func__);

	par->display->fbtftops.reset(par);

	/* startup sequence for MI0283QT-9A */
	write_reg(par, MIPI_DCS_SOFT_RESET);
	mdelay(5);
	write_reg(par, MIPI_DCS_SET_DISPLAY_OFF);
	/* --------------------------------------------------------- */
	write_reg(par, 0xCF, 0x00, 0x83, 0x30);
	write_reg(par, 0xED, 0x64, 0x03, 0x12, 0x81);
	write_reg(par, 0xE8, 0x85, 0x01, 0x79);
	write_reg(par, 0xCB, 0x39, 0X2C, 0x00, 0x34, 0x02);
	write_reg(par, 0xF7, 0x20);
	write_reg(par, 0xEA, 0x00, 0x00);
	/* ------------power control-------------------------------- */
	write_reg(par, 0xC0, 0x26);
	write_reg(par, 0xC1, 0x11);
	/* ------------VCOM --------- */
	write_reg(par, 0xC5, 0x35, 0x3E);
	write_reg(par, 0xC7, 0xBE);
	/* ------------memory access control------------------------ */
	write_reg(par, MIPI_DCS_SET_PIXEL_FORMAT, 0x55); /* 16bit pixel */
	/* ------------frame rate----------------------------------- */
	write_reg(par, 0xB1, 0x00, 0x1B);
	/* ------------Gamma---------------------------------------- */
	/* write_reg(par, 0xF2, 0x08); */ /* Gamma Function Disable */
	write_reg(par, MIPI_DCS_SET_GAMMA_CURVE, 0x01);
	/* ------------display-------------------------------------- */
	write_reg(par, 0xB7, 0x07); /* entry mode set */
	write_reg(par, 0xB6, 0x0A, 0x82, 0x27, 0x00);
	write_reg(par, MIPI_DCS_EXIT_SLEEP_MODE);
	mdelay(100);
	write_reg(par, MIPI_DCS_SET_DISPLAY_ON);
	mdelay(20);

	return 0;
}

#define MEM_Y   BIT(7) /* MY row address order */
#define MEM_X   BIT(6) /* MX column address order */
#define MEM_V   BIT(5) /* MV row / column exchange */
#define MEM_L   BIT(4) /* ML vertical refresh order */
#define MEM_H   BIT(2) /* MH horizontal refresh order */
#define MEM_BGR (3) /* RGB-BGR Order */
static int set_var(struct fbtft_par *par)
{
	fbtft_dbg("%s()\n", __func__);

	switch (par->rotate) {
	case 0:
		write_reg(par, MIPI_DCS_SET_ADDRESS_MODE,
			  MEM_X | (par->bgr << MEM_BGR));
		break;
	case 270:
		write_reg(par, MIPI_DCS_SET_ADDRESS_MODE,
			  MEM_V | MEM_L | (par->bgr << MEM_BGR));
		break;
	case 180:
		write_reg(par, MIPI_DCS_SET_ADDRESS_MODE,
			  MEM_Y | (par->bgr << MEM_BGR));
		break;
	case 90:
		write_reg(par, MIPI_DCS_SET_ADDRESS_MODE,
			  MEM_Y | MEM_X | MEM_V | (par->bgr << MEM_BGR));
		break;
	}

	return 0;
}

/*
 * Gamma string format:
 *  Positive: Par1 Par2 [...] Par15
 *  Negative: Par1 Par2 [...] Par15
 */
#define CURVE(num, idx)  par->display->gamma.curves[(num) * \
				par->display->gamma.num_values + (idx)]
static int set_gamma(struct fbtft_par *par)
{
	int i;

	fbtft_dbg("%s()\n", __func__);

	for (i = 0; i < par->display->gamma.num_curves; i++)
		write_reg(par, 0xE0 + i,
			  CURVE(i, 0), CURVE(i, 1), CURVE(i, 2),
			  CURVE(i, 3), CURVE(i, 4), CURVE(i, 5),
			  CURVE(i, 6), CURVE(i, 7), CURVE(i, 8),
			  CURVE(i, 9), CURVE(i, 10), CURVE(i, 11),
			  CURVE(i, 12), CURVE(i, 13), CURVE(i, 14));

	return 0;
}

#undef CURVE

struct fbtft_display ili9341 = {
	.regwidth = 8,
	.buswidth = FBTFT_BUSWIDTH,
	.width = WIDTH,
	.height = HEIGHT,
	.bpp = (1 << (LCD_BPP)),
	.gamma = {
		.num_curves = 2,
		.num_values = 15,
		.curves = DEFAULT_GAMMA,
	},
	.fbtftops = {
		.init_display = init_display,
		.set_var = set_var,
		.set_gamma = set_gamma,
	},
};

struct fbtft_par par = {
	.spi_bus = FBTFT_SPI_BUS,
	.spi_cs = FBTFT_SPI_CS,
	.spi_max_hz = FBTFT_SPI_MAX_HZ,
	.display = &ili9341,
	.gpio = {
			.reset = FBTFT_GPIO_RESET,
			.dc = FBTFT_GPIO_DC,
		},
	.bgr = FBTFT_BGR,
	.rotate = FBTFT_ROTATE,
};

vidinfo_t panel_info = {
#if (FBTFT_ROTATE == 90) || (FBTFT_ROTATE == 270)
	.vl_row  = WIDTH,
	.vl_col  = HEIGHT,
#else
	.vl_col  = WIDTH,
	.vl_row  = HEIGHT,
#endif
	.vl_bpix = LCD_BPP,       /* Bits per pixel */
	.cmap = NULL,             /* color map */
	.priv = &par,             /* driver-specific data */
};
