/*
 * Copyright 2016 Broadcom
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation (the "GPL").
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License version 2 (GPLv2) for more details.
 *
 * You should have received a copy of the GNU General Public License
 * version 2 (GPLv2) along with this source code.
 */

#include <linux/compiler.h>
#include <linux/compat.h>
#include <asm/types.h>
#include <asm/system.h>
#include <config.h>
#include <common.h>
#include <malloc.h>
#include <asm/io.h>
#include <asm/dma-mapping.h>

#include <asm/arch/bcm_iproc_smc.h>



struct secure_bridge_data {
	void __iomem *bounce;		/* virtual address */
	u32 buffer_addr;				/* physical address */
	int initialized;
} bridge_data;

struct bcm_iproc_smc_data {
	unsigned service_id;
	unsigned arg0;
	unsigned arg1;
	unsigned arg2;
	unsigned arg3;
	int ret_val;
};



static u32 bcm_iproc_smc_asm(unsigned service_id)
{
	/* Set Up Registers to pass data to Secure Monitor */
	register u32 r4 asm("r4") = service_id;
	register u32 r5 asm("r5") = 0x3;	/* Keep IRQ and FIQ off in SM */
	register u32 r6 asm("r6") = bridge_data.buffer_addr;
	register u32 r12 asm("r12");
	register u32 r0 asm("r0");

	do {
		/* Secure Monitor Call */
		asm volatile (__asmeq("%0", "ip")
	      __asmeq("%1", "r4")
	      __asmeq("%2", "r5") __asmeq("%3", "r6")
	      ".arch_extension sec\n"
	      "smc	#0		@ switch to secure world\n" :
	      "+r"(r12), "+r"(r4), "+r"(r5), "+r"(r6), "+r"(r0)
	       : : "r1", "r2", "r3", "r7", "r8", "r14");

		/* Setup registers in case we need to re-enter secure mode */
		r4 = SSAPI_RET_FROM_INT_SERV;
		r5 = 0x3;
		r6 = bridge_data.buffer_addr;

	} while (r12 != SEC_EXIT_NORMAL);

	return r0;
}

/* __bcm_iproc_smc() should only run on CPU 0, with pre-emption disabled */
static void __bcm_iproc_smc(void *info)
{
	struct bcm_iproc_smc_data *data = info;
	u32 *args = bridge_data.bounce;
	int rc = 0;

	/* Must run on CPU 0 */

	/* Check map in the bounce area */
	BUG_ON(!bridge_data.initialized);

	/* Copy one 32 bit word into the bounce area */

	args[0] = data->arg0;
	args[1] = data->arg1;
	args[2] = data->arg2;
	args[3] = data->arg3;

	/* Prior to switching to secure mode, it is essential
	 * to flush the cache and update all data into DDR
	 * memory
	 */
	flush_dcache_all();

	/* Trap into Secure Monitor */
	rc = bcm_iproc_smc_asm(data->service_id);
	if (rc != SEC_ROM_RET_OK)
		printf("Secure Monitor call failed (0x%x)!\n", rc);
	/* forward the smc return value to calling function */
	data->ret_val = rc;
}

unsigned bcm_iproc_smc(unsigned service_id, unsigned arg0, unsigned arg1,
		  unsigned arg2, unsigned arg3)
{
	struct bcm_iproc_smc_data data;

	data.service_id = service_id;
	data.arg0 = arg0;
	data.arg1 = arg1;
	data.arg2 = arg2;
	data.arg3 = arg3;
	/*
	 * In the case of SMP environment, due to a limitation of the
	 * secure monitor design, we must use the SMP
	 * infrastructure to forward all secure monitor calls to Core 0.
	 */
	 __bcm_iproc_smc((void *)&data);

	return data.ret_val;
}


int bcm_ssapi_call(u32 service_id, u32 arg0, u32 arg1,
	u32 arg2, u32 arg3, u32 arg4, u32 arg5,
	u32 arg6, u32 arg7, u32 arg8)
{
	int ret = 1;
	/*strucure virurall address and phy  */
	int st_size;
	int hlen = sizeof(struct bcm_iproc_smc_data);

	void *st_vir = NULL;
	dma_addr_t st_phy;

	/*in/output buffer vitural address and phy  */
	int dma_size = 0;
	void *dma_vir = NULL;
	dma_addr_t dma_phy;

	struct ssapi_hash_create_data *p_hash;
	struct ssapi_public_decrypt_data *p_de;

	switch (service_id) {
	case SSAPI_BRCM_HASH_CREATE:

		st_size = sizeof(struct ssapi_hash_create_data);
		st_vir = dma_alloc_coherent(hlen + st_size,
		&st_phy);

		if (!st_vir) {
			printf("st_vir addr alloc failed\n");
			ret = -ENOMEM;
		}
		memset(st_vir, 0, hlen + st_size);

		p_hash = (struct ssapi_hash_create_data *)(st_vir + hlen);
		p_hash->output_size = arg2;
		p_hash->input_size = arg4;
		p_hash->hashAlgo = arg5;
		p_hash->useDma = arg6;
		p_hash->input_data_phy = arg3;

		bridge_data.initialized = 1;
		bridge_data.bounce = st_vir;
		bridge_data.buffer_addr =  st_phy;

		bcm_iproc_smc(service_id,
			p_hash,
			st_size,
			&p_hash->output_data_phy[0],
			p_hash->output_size);

		memcpy(arg1, p_hash->output_data_phy, 32);

	break;
	case SSAPI_BRCM_RSA_PUBLIC_DECRYPT:

		st_size = sizeof(struct ssapi_public_decrypt_data);
		st_vir = dma_alloc_coherent(hlen + st_size, &st_phy);

		if (!st_vir) {
			printf("st_vir addr alloc failed\n");
			ret = -ENOMEM;
		}
		memset(st_vir, 0, hlen + st_size);
		p_de = (struct ssapi_public_decrypt_data *)(st_vir + hlen);

		p_de->decryptedDataSize = 256;
		p_de->encryptedDataSize = arg4;
		p_de->modLen = arg6;
		p_de->exp = arg7;

		/* Allocate memory for the buffers of crypto service */
		dma_size += p_de->encryptedDataSize;
		dma_size += p_de->modLen;
		dma_size += p_de->decryptedDataSize;

		if (dma_size > 0) {
			dma_vir = dma_alloc_coherent(dma_size, &dma_phy);
			if (!dma_vir) {
				printf("dma buffers alloc for ssapi_public_decrypt_data failed\n");
				ret = -ENOMEM;
			}
		}

		memset(dma_vir, 0, dma_size);

		p_de->encryptedData_phy = dma_phy;
		p_de->keyLocation_phy = dma_phy + p_de->encryptedDataSize;
		p_de->decryptedData_phy = p_de->keyLocation_phy + p_de->modLen;


		/* Update the Input to Crypto operation */
		memcpy(p_de->encryptedData_phy, arg3, p_de->encryptedDataSize);

		/* Update the key to */
		memcpy(p_de->keyLocation_phy, arg5, p_de->modLen);

		/*  clear the decrypted-outputbuf  */
		memset(p_de->decryptedData_phy, 0, p_de->decryptedDataSize);

		bridge_data.initialized = 1;
		bridge_data.bounce = st_vir;
		bridge_data.buffer_addr =  st_phy;

		bcm_iproc_smc(service_id, p_de, st_size,
			p_de->decryptedData_phy, p_de->decryptedDataSize);

		memcpy((unsigned char *)arg1,
			(unsigned char *)p_de->decryptedData_phy,
			p_de->decryptedDataSize);

		*(unsigned char *)(arg2) = p_de->decryptedDataSize;

	break;
	default:
	return ret;
	}
	if (dma_vir)
		free(dma_vir);

	if (st_vir)
		free(st_vir);

	return ret;
}
