/*
 *
 * Copyright (C) 2016 Broadcom
 *
 * SPDX-License-Identifier:      GPL-2.0+
 *
 */

#include <common.h>
#include <asm/gpio.h>
#include <asm/io.h>

#ifdef CYGNUS_GPIO_DEBUG
#define gpio_dbg(format, arg...) printf(format, ##arg)
#else
#define gpio_dbg(format, arg...)
#endif

#define BIT(x) (1<<(x))

#define CYGNUS_GPIO_CTRL_NB 3
#define CYGNUS_MAX_GPIO 176

#define CYGNUS_GPIO_DATA_IN_OFFSET   0x00
#define CYGNUS_GPIO_DATA_OUT_OFFSET  0x04
#define CYGNUS_GPIO_OUT_EN_OFFSET    0x08

#define GPIO_BANK_SIZE 0x200
#define NGPIOS_PER_BANK 32
#define GPIO_BANK(pin) ((pin) / NGPIOS_PER_BANK)

#define CYGNUS_GPIO_REG(pin, reg) (GPIO_BANK(pin) * GPIO_BANK_SIZE + (reg))
#define CYGNUS_GPIO_SHIFT(pin) ((pin) % NGPIOS_PER_BANK)

/*
 * @offset: register offset for GPIO mux override of a pin
 * @shift: bit shift for GPIO mux override of a pin
 * @override: has gpio mux override
 */
struct cygnus_gpio_mux {
	unsigned char override;
	unsigned char offset;
	unsigned char shift;
};

#define CYGNUS_GPIO_MUX_DESC(num, name, m, o, s)	\
	/* gpio# num  pad name */			\
	{						\
		.override = m,				\
		.offset = o,				\
		.shift = s,				\
	}						\

/*
 * Cygnux asiu gpio mux
 */
static struct cygnus_gpio_mux gpio_asiu_mux[] = {
	CYGNUS_GPIO_MUX_DESC(0, "sc0_clk", 1, 0x10, 22),
	CYGNUS_GPIO_MUX_DESC(1, "sc0_detect", 1, 0x10, 18),
	CYGNUS_GPIO_MUX_DESC(2, "sc0_fcb", 1, 0x10, 16),
	CYGNUS_GPIO_MUX_DESC(3, "sc0_io", 1, 0x10, 14),
	CYGNUS_GPIO_MUX_DESC(4, "sc1_clk", 1, 0x10, 10),
	CYGNUS_GPIO_MUX_DESC(5, "sc1_detect", 1, 0x10, 6),
	CYGNUS_GPIO_MUX_DESC(6, "sc1_fcb", 1, 0x10, 4),
	CYGNUS_GPIO_MUX_DESC(7, "sc1_io", 1, 0x10, 2),
	CYGNUS_GPIO_MUX_DESC(8, "lcd_clac", 1, 0x4, 26),
	CYGNUS_GPIO_MUX_DESC(9, "lcd_clpower", 1, 0x8, 0),
	CYGNUS_GPIO_MUX_DESC(10, "lcd_clfp", 1, 0x8, 6),
	CYGNUS_GPIO_MUX_DESC(11, "lcd_cllp", 1, 0x8, 2),
	CYGNUS_GPIO_MUX_DESC(12, "lcd_clle", 1, 0x8, 4),
	CYGNUS_GPIO_MUX_DESC(13, "lcd_clcp", 1, 0x4, 24),
	CYGNUS_GPIO_MUX_DESC(14, "lcd_cld0", 1, 0x4, 22),
	CYGNUS_GPIO_MUX_DESC(15, "lcd_cld1", 1, 0x4, 0),
	CYGNUS_GPIO_MUX_DESC(16, "lcd_cld2", 1, 0x8, 22),
	CYGNUS_GPIO_MUX_DESC(17, "lcd_cld3", 1, 0x8, 20),
	CYGNUS_GPIO_MUX_DESC(18, "lcd_cld4", 1, 0x8, 18),
	CYGNUS_GPIO_MUX_DESC(19, "lcd_cld5", 1, 0x8, 16),
	CYGNUS_GPIO_MUX_DESC(20, "lcd_cld6", 1, 0x8, 14),
	CYGNUS_GPIO_MUX_DESC(21, "lcd_cld7", 1, 0x8, 12),
	CYGNUS_GPIO_MUX_DESC(22, "lcd_cld8", 1, 0x8, 10),
	CYGNUS_GPIO_MUX_DESC(23, "lcd_cld9", 1, 0x8, 8),
	CYGNUS_GPIO_MUX_DESC(24, "lcd_cld10", 1, 0x4, 20),
	CYGNUS_GPIO_MUX_DESC(25, "lcd_cld11", 1, 0x4, 18),
	CYGNUS_GPIO_MUX_DESC(26, "lcd_cld12", 1, 0x4, 16),
	CYGNUS_GPIO_MUX_DESC(27, "lcd_cld13", 1, 0x4, 14),
	CYGNUS_GPIO_MUX_DESC(28, "lcd_cld14", 1, 0x4, 12),
	CYGNUS_GPIO_MUX_DESC(29, "lcd_cld15", 1, 0x4, 10),
	CYGNUS_GPIO_MUX_DESC(30, "lcd_cld16", 1, 0x4, 8),
	CYGNUS_GPIO_MUX_DESC(31, "lcd_cld17", 1, 0x4, 6),
	CYGNUS_GPIO_MUX_DESC(32, "lcd_cld18", 1, 0x4, 4),
	CYGNUS_GPIO_MUX_DESC(33, "lcd_cld19", 1, 0x4, 2),
	CYGNUS_GPIO_MUX_DESC(34, "lcd_cld20", 1, 0x8, 30),
	CYGNUS_GPIO_MUX_DESC(35, "lcd_cld21", 1, 0x8, 28),
	CYGNUS_GPIO_MUX_DESC(36, "lcd_cld22", 1, 0x8, 26),
	CYGNUS_GPIO_MUX_DESC(37, "lcd_cld23", 1, 0x8, 24),
	CYGNUS_GPIO_MUX_DESC(38, "spi0_clk", 1, 0x18, 10),
	CYGNUS_GPIO_MUX_DESC(39, "spi0_miso", 1, 0x18, 8),
	CYGNUS_GPIO_MUX_DESC(40, "spi0_ss", 1, 0x18, 4),
	CYGNUS_GPIO_MUX_DESC(41, "spi1_clk", 1, 0x18, 2),
	CYGNUS_GPIO_MUX_DESC(42, "spi1_miso", 1, 0x18, 0),
	CYGNUS_GPIO_MUX_DESC(43, "spi1_ss", 1, 0x1c, 28),
	CYGNUS_GPIO_MUX_DESC(44, "spi2_clk", 1, 0x1c, 26),
	CYGNUS_GPIO_MUX_DESC(45, "spi2_miso", 1, 0x1c, 24),
	CYGNUS_GPIO_MUX_DESC(46, "spi2_ss", 1, 0x1c, 20),
	CYGNUS_GPIO_MUX_DESC(47, "spi3_clk", 1, 0x1c, 18),
	CYGNUS_GPIO_MUX_DESC(48, "spi3_miso", 1, 0x1c, 16),
	CYGNUS_GPIO_MUX_DESC(49, "spi3_ss", 1, 0x1c, 12),
	CYGNUS_GPIO_MUX_DESC(50, "sdio1_clk_sdcard", 1, 0x14, 6),
	CYGNUS_GPIO_MUX_DESC(51, "sdio1_cmd", 1, 0x14, 4),
	CYGNUS_GPIO_MUX_DESC(52, "sdio1_data0", 1, 0x14, 2),
	CYGNUS_GPIO_MUX_DESC(53, "sdio1_data1", 1, 0x14, 0),
	CYGNUS_GPIO_MUX_DESC(54, "sdio1_data2", 1, 0x18, 30),
	CYGNUS_GPIO_MUX_DESC(55, "sdio1_data3", 1, 0x18, 28),
	CYGNUS_GPIO_MUX_DESC(56, "sdio0_clk_sdcard", 1, 0x14, 26),
	CYGNUS_GPIO_MUX_DESC(57, "sdio0_cmd", 1, 0x14, 24),
	CYGNUS_GPIO_MUX_DESC(58, "sdio0_data0", 1, 0x14, 22),
	CYGNUS_GPIO_MUX_DESC(59, "sdio0_data1", 1, 0x14, 20),
	CYGNUS_GPIO_MUX_DESC(60, "sdio0_data2", 1, 0x14, 18),
	CYGNUS_GPIO_MUX_DESC(61, "sdio0_data3", 1, 0x14, 16),
	CYGNUS_GPIO_MUX_DESC(62, "uart0_rx", 1, 0x1c, 6),
	CYGNUS_GPIO_MUX_DESC(63, "uart0_cts", 1, 0x1c, 10),
	CYGNUS_GPIO_MUX_DESC(64, "uart1_rx", 1, 0x20, 6),
	CYGNUS_GPIO_MUX_DESC(65, "uart1_cts", 1, 0x1c, 2),
	CYGNUS_GPIO_MUX_DESC(66, "uart1_dcd", 1, 0x1c, 0),
	CYGNUS_GPIO_MUX_DESC(67, "uart1_dsr", 1, 0x20, 14),
	CYGNUS_GPIO_MUX_DESC(68, "uart1_ri", 1, 0x20, 10),
	CYGNUS_GPIO_MUX_DESC(69, "uart3_rx", 1, 0x20, 2),
	CYGNUS_GPIO_MUX_DESC(70, "camera_vsync", 1, 0x4, 30),
	CYGNUS_GPIO_MUX_DESC(71, "camera_trigger", 1, 0x0, 0),
	CYGNUS_GPIO_MUX_DESC(72, "camera_strobe", 1, 0x0, 2),
	CYGNUS_GPIO_MUX_DESC(73, "camera_standby", 1, 0x0, 4),
	CYGNUS_GPIO_MUX_DESC(74, "camera_reset_n", 1, 0x0, 6),
	CYGNUS_GPIO_MUX_DESC(75, "camera_pixdata9", 1, 0x0, 8),
	CYGNUS_GPIO_MUX_DESC(76, "camera_pixdata8", 1, 0x0, 10),
	CYGNUS_GPIO_MUX_DESC(77, "camera_pixdata7", 1, 0x0, 12),
	CYGNUS_GPIO_MUX_DESC(78, "camera_pixdata6", 1, 0x0, 14),
	CYGNUS_GPIO_MUX_DESC(79, "camera_pixdata5", 1, 0x0, 16),
	CYGNUS_GPIO_MUX_DESC(80, "camera_pixdata4", 1, 0x0, 18),
	CYGNUS_GPIO_MUX_DESC(81, "camera_pixdata3", 1, 0x0, 20),
	CYGNUS_GPIO_MUX_DESC(82, "camera_pixdata2", 1, 0x0, 22),
	CYGNUS_GPIO_MUX_DESC(83, "camera_pixdata1", 1, 0x0, 24),
	CYGNUS_GPIO_MUX_DESC(84, "camera_pixdata0", 1, 0x0, 26),
	CYGNUS_GPIO_MUX_DESC(85, "camera_pixclk", 1, 0x0, 28),
	CYGNUS_GPIO_MUX_DESC(86, "camera_hsync", 1, 0x0, 30),
	CYGNUS_GPIO_MUX_DESC(87, "sflash_clk", 1, 0x18, 22),
	CYGNUS_GPIO_MUX_DESC(88, "sflash_cs_l", 1, 0x18, 20),
	CYGNUS_GPIO_MUX_DESC(89, "sflash_mosi", 1, 0x18, 14),
	CYGNUS_GPIO_MUX_DESC(90, "sflash_miso", 1, 0x18, 16),
	CYGNUS_GPIO_MUX_DESC(91, "sflash_wp_n", 1, 0x18, 12),
	CYGNUS_GPIO_MUX_DESC(92, "sflash_hold_n", 1, 0x18, 18),
	CYGNUS_GPIO_MUX_DESC(93, "nand_ale", 1, 0xc, 30),
	CYGNUS_GPIO_MUX_DESC(94, "nand_ce0_l", 1, 0xc, 28),
	CYGNUS_GPIO_MUX_DESC(95, "nand_ce1_l", 1, 0xc, 26),
	CYGNUS_GPIO_MUX_DESC(96, "nand_cle", 1, 0xc, 24),
	CYGNUS_GPIO_MUX_DESC(97, "nand_dq0", 1, 0xc, 22),
	CYGNUS_GPIO_MUX_DESC(98, "nand_dq1", 1, 0xc, 20),
	CYGNUS_GPIO_MUX_DESC(99, "sdio0_led_on", 1, 0x14, 14),
	CYGNUS_GPIO_MUX_DESC(100, "sdio0_cd", 1, 0x14, 28),
	CYGNUS_GPIO_MUX_DESC(101, "sdio1_wp_n", 1, 0x18, 24),
	CYGNUS_GPIO_MUX_DESC(102, "sdio1_card_rst", 1, 0x14, 10),
	CYGNUS_GPIO_MUX_DESC(103, "sdio1_led_on", 1, 0x18, 26),
	CYGNUS_GPIO_MUX_DESC(104, "sdio1_cd", 1, 0x14, 8),
	CYGNUS_GPIO_MUX_DESC(105, "nand_dq2", 1, 0xc, 18),
	CYGNUS_GPIO_MUX_DESC(106, "nand_dq3", 1, 0xc, 16),
	CYGNUS_GPIO_MUX_DESC(107, "nand_dq4", 1, 0xc, 14),
	CYGNUS_GPIO_MUX_DESC(108, "nand_dq5", 1, 0xc, 12),
	CYGNUS_GPIO_MUX_DESC(109, "nand_dq6", 1, 0xc, 10),
	CYGNUS_GPIO_MUX_DESC(110, "nand_dq7", 1, 0xc, 8),
	CYGNUS_GPIO_MUX_DESC(111, "sdio0_wp_n", 1, 0x14, 12),
	CYGNUS_GPIO_MUX_DESC(112, "sdio0_card_rst", 1, 0x14, 30),
	CYGNUS_GPIO_MUX_DESC(113, "nand_rb_l", 1, 0xc, 6),
	CYGNUS_GPIO_MUX_DESC(114, "nand_re_l", 1, 0xc, 4),
	CYGNUS_GPIO_MUX_DESC(115, "nand_we_l", 1, 0xc, 2),
	CYGNUS_GPIO_MUX_DESC(116, "nand_wp_l", 1, 0xc, 0),

	CYGNUS_GPIO_MUX_DESC(117, "", 0, 0, 0),
	CYGNUS_GPIO_MUX_DESC(118, "", 0, 0, 0),
	CYGNUS_GPIO_MUX_DESC(119, "", 0, 0, 0),
	CYGNUS_GPIO_MUX_DESC(120, "", 0, 0, 0),
	CYGNUS_GPIO_MUX_DESC(121, "", 0, 0, 0),
	CYGNUS_GPIO_MUX_DESC(122, "", 0, 0, 0),

	CYGNUS_GPIO_MUX_DESC(123, "d1wowstz_l", 1, 0x4, 28),
	CYGNUS_GPIO_MUX_DESC(124, "pwm0", 1, 0x10, 30),
	CYGNUS_GPIO_MUX_DESC(125, "pwm1", 1, 0x10, 28),
	CYGNUS_GPIO_MUX_DESC(126, "pwm2", 1, 0x10, 26),
	CYGNUS_GPIO_MUX_DESC(127, "pwm3", 1, 0x10, 24),
	CYGNUS_GPIO_MUX_DESC(128, "sc0_cmdvcc_l", 1, 0x10, 20),
	CYGNUS_GPIO_MUX_DESC(129, "sc0_rst_l", 1, 0x10, 12),
	CYGNUS_GPIO_MUX_DESC(130, "sc1_cmdvcc_l", 1, 0x10, 8),
	CYGNUS_GPIO_MUX_DESC(131, "sc1_rst_l", 1, 0x10, 0),
	CYGNUS_GPIO_MUX_DESC(132, "spi0_mosi", 1, 0x18, 6),
	CYGNUS_GPIO_MUX_DESC(133, "spi1_mosi", 1, 0x1c, 30),
	CYGNUS_GPIO_MUX_DESC(134, "spi2_mosi", 1, 0x1c, 22),
	CYGNUS_GPIO_MUX_DESC(135, "spi3_mosi", 1, 0x1c, 14),
	CYGNUS_GPIO_MUX_DESC(136, "uart0_rts", 1, 0x1c, 8),
	CYGNUS_GPIO_MUX_DESC(137, "uart0_tx", 1, 0x1c, 4),
	CYGNUS_GPIO_MUX_DESC(138, "uart1_dtr", 1, 0x20, 12),
	CYGNUS_GPIO_MUX_DESC(139, "uart1_rts", 1, 0x20, 8),
	CYGNUS_GPIO_MUX_DESC(140, "uart1_tx", 1, 0x20, 4),
	CYGNUS_GPIO_MUX_DESC(141, "uart3_tx", 1, 0x20, 0),
	CYGNUS_GPIO_MUX_DESC(142, "d1w_dq", 1, 0x28, 0),

	CYGNUS_GPIO_MUX_DESC(143, "", 0, 0, 0),
	CYGNUS_GPIO_MUX_DESC(144, "", 0, 0, 0),
	CYGNUS_GPIO_MUX_DESC(145, "", 0, 0, 0),
};

/*
 * Cygnus GPIO controller
 *
 * @base: I/O register base for Cygnus GPIO controller
 * @io_ctrl: I/O register base for Cygnus GPIO PINCTRL
 * @start: start gpio number
 * @count: number of gpio
 */
struct cygnus_gpio {
	void __iomem *base;
	void __iomem *io_ctrl;
	struct cygnus_gpio_mux *mux;
	unsigned start;
	unsigned count;
};

struct cygnus_gpio ctrl[CYGNUS_GPIO_CTRL_NB] = {
	{
		/* gpio-ccm */
		.base = (void *)0x1800a000,
		.start = 0,
		.count = 24,
	},
	{
		/* gpio-asiu */
		.base = (void *)0x180a5000,
		.io_ctrl = (void *)0x0301d24c,
		.mux = gpio_asiu_mux,
		.start = 24,
		.count = 146,
	},
	{
		/* gpio-crmu */
		.base = (void *)0x03024800,
		.start = 170,
		.count = 6,
	},
};

static struct cygnus_gpio *cygnus_gpio_map(unsigned gpio)
{
	int i = 0;
	while (i < CYGNUS_GPIO_CTRL_NB) {
		if ((gpio >= ctrl[i].start) &&
		    (gpio < (ctrl[i].start + ctrl[i].count))) {
			return &ctrl[i];
		}
		i++;
	}
	printf("%s: invalid gpio %d\n", __func__, gpio);
	return NULL;
}

/**
 *  cygnus_set_bit - set or clear one bit (corresponding to the GPIO pin) in a
 *  Cygnus GPIO register
 *
 *  @cygnus_gpio: Cygnus GPIO device
 *  @reg: register offset
 *  @gpio: GPIO pin
 *  @set: set or clear
 */
static inline void cygnus_set_bit(struct cygnus_gpio *chip, unsigned int reg,
				  unsigned gpio, bool set)
{
	unsigned int offset = CYGNUS_GPIO_REG(gpio, reg);
	unsigned int shift = CYGNUS_GPIO_SHIFT(gpio);
	u32 val;

	val = readl(chip->base + offset);
	if (set)
		val |= BIT(shift);
	else
		val &= ~BIT(shift);
	writel(val, chip->base + offset);
}

static inline bool cygnus_get_bit(struct cygnus_gpio *chip, unsigned int reg,
			  unsigned gpio)
{
	unsigned int offset = CYGNUS_GPIO_REG(gpio, reg);
	unsigned int shift = CYGNUS_GPIO_SHIFT(gpio);

	return !!(readl(chip->base + offset) & BIT(shift));
}

static void cygnus_pinctrl_request_gpio(struct cygnus_gpio *chip,
					unsigned offset)
{
	u32 val;
	void *reg;

	if (!chip->mux[offset].override)
		return;

	reg = chip->io_ctrl + chip->mux[offset].offset;
	val = readl(reg);
	val |= 0x3 << chip->mux[offset].shift;
	writel(val, reg);

	gpio_dbg("%s: reg=0x%x shift=%u\n", __func__,
		 (unsigned int)reg, chip->mux[offset].shift);
}

int gpio_is_valid(int gpio)
{
	return (gpio >= 0) && (gpio < CYGNUS_MAX_GPIO);
}

int gpio_request(unsigned gpio, const char *label)
{
	struct cygnus_gpio *chip = cygnus_gpio_map(gpio);
	gpio_dbg("gpio:%u request\n", gpio);
	if (!chip)
		return -1;

	if (!chip->io_ctrl)
		return 0;

	cygnus_pinctrl_request_gpio(chip, gpio - chip->start);
	return 0;
}

int gpio_free(unsigned gpio)
{
	struct cygnus_gpio *chip = cygnus_gpio_map(gpio);

	gpio_dbg("gpio:%u free\n", gpio);
	if (!chip)
		return -1;

	return 0;
}

int gpio_direction_input(unsigned gpio)
{
	struct cygnus_gpio *chip = cygnus_gpio_map(gpio);

	gpio_dbg("gpio:%u set input\n", gpio);
	if (!chip)
		return -1;

	cygnus_set_bit(chip, CYGNUS_GPIO_OUT_EN_OFFSET, gpio - chip->start,
		       false);

	return 0;
}

int gpio_direction_output(unsigned gpio, int val)
{
	struct cygnus_gpio *chip = cygnus_gpio_map(gpio);

	gpio_dbg("gpio:%u set output, value:%d\n", gpio, val);
	if (!chip)
		return -1;

	cygnus_set_bit(chip, CYGNUS_GPIO_DATA_OUT_OFFSET, gpio - chip->start,
		       !!(val));
	cygnus_set_bit(chip, CYGNUS_GPIO_OUT_EN_OFFSET, gpio - chip->start,
		       true);

	return 0;
}

int gpio_set_value(unsigned gpio, int val)
{
	struct cygnus_gpio *chip = cygnus_gpio_map(gpio);

	gpio_dbg("gpio:%u set, value:%d\n", gpio, val);
	if (!chip)
		return -1;

	cygnus_set_bit(chip, CYGNUS_GPIO_DATA_OUT_OFFSET, gpio - chip->start,
		       !!(val));

	return 0;
}

int gpio_get_value(unsigned gpio)
{
	int val;
	struct cygnus_gpio *chip = cygnus_gpio_map(gpio);

	gpio_dbg("gpio:%u get", gpio);
	if (!chip)
		return -1;

	val = cygnus_get_bit(chip, CYGNUS_GPIO_DATA_IN_OFFSET,
			     gpio - chip->start);
	gpio_dbg(", value:%d\n", val);
	return val;
}
