/*
 * Copyright (C) 2013 Noralf Tronnes
 * Copyright (C) 2016 Broadcom
 *
 * SPDX-License-Identifier:      GPL-2.0+
 *
 */

#include <common.h>
#include <asm/gpio.h>
#include <spi.h>
#include <lcd.h>
#include "mipi_display.h"
#include "fbtft_core.h"

/*****************************************************************************
 *
 *   void (*write_reg)(struct fbtft_par *par, int len, ...);
 *
 *****************************************************************************/

#define no_conversion
#define define_fbtft_write_reg(func, type, modifier)                          \
static void func(struct fbtft_par *par, int len, ...)                         \
{                                                                             \
	va_list args;                                                         \
	int i, ret;                                                           \
	int offset = 0;                                                       \
	type *buf = (type *)par->buf;                                         \
									      \
	fbtft_dbg("%s(len=%zu)\n", __func__, len);                            \
									      \
	va_start(args, len);                                                  \
									      \
	if (par->startbyte) {                                                 \
		*(u8 *)par->buf = par->startbyte;                             \
		buf = (type *)(par->buf + 1);                                 \
		offset = 1;                                                   \
	}                                                                     \
									      \
	*buf = modifier((type)va_arg(args, unsigned int));                    \
	if (par->gpio.dc != -1)                                               \
		gpio_set_value(par->gpio.dc, 0);                              \
	ret = par->display->fbtftops.write(par, par->buf,                     \
					   sizeof(type) + offset);            \
	if (ret < 0) {                                                        \
		va_end(args);                                                 \
		error("write failed and returned %d\n", ret);                 \
		return;                                                       \
	}                                                                     \
	len--;                                                                \
									      \
	if (par->startbyte)                                                   \
		*(u8 *)par->buf = par->startbyte | 0x2;                       \
									      \
	if (len) {                                                            \
		i = len;                                                      \
		while (i--) {                                                 \
			*buf++ = modifier((type)va_arg(args, unsigned int));  \
		}                                                             \
		if (par->gpio.dc != -1)                                       \
			gpio_set_value(par->gpio.dc, 1);                      \
		ret = par->display->fbtftops.write(par, par->buf,             \
						   len *                      \
						   (sizeof(type) + offset));  \
		if (ret < 0) {                                                \
			va_end(args);                                         \
			error("write failed and returned %d\n", ret);         \
			return;                                               \
		}                                                             \
	}                                                                     \
	va_end(args);                                                         \
}

define_fbtft_write_reg(fbtft_write_reg8_bus8, u8, no_conversion)

/*****************************************************************************
 *
 *   int (*write_vmem)(struct fbtft_par *par);
 *
 *****************************************************************************/

/* 16 bit pixel over 8-bit databus */
static int fbtft_write_vmem16_bus8(struct fbtft_par *par, size_t offset,
				   size_t len)
{
	u16 *vmem16;

	fbtft_dbg("%s(offset=%zu, len=%zu)\n", __func__, offset, len);

	vmem16 = (u16 *)(par->fb + offset);

	if (par->gpio.dc != -1)
		gpio_set_value(par->gpio.dc, 1);

	return par->display->fbtftops.write(par, vmem16, len);
}

static int fbtft_write_spi(struct fbtft_par *par, void *buf, size_t len)
{
	int ret;

	fbtft_dbg_hex(u8, buf, len, "%s(len=%d): ", __func__, len);

	if (spi_claim_bus(par->slave)) {
		error("failed to claim SPI bus\n");
		return -1;
	}

	ret = spi_xfer(par->slave, len * 8, buf, NULL,
			SPI_XFER_BEGIN | SPI_XFER_END);

	if (ret)
		error("SPI xfer failed %d\n", ret);

	spi_release_bus(par->slave);

	return ret;
}

static void fbtft_set_addr_win(struct fbtft_par *par, int xs, int ys, int xe,
			       int ye)
{
	fbtft_dbg("%s(xs=%zu, ys=%zu, xe=%zu, ye=%zu)\n", __func__,
		  xs, ys, xe, ye);

	write_reg(par, MIPI_DCS_SET_COLUMN_ADDRESS,
		  (xs >> 8) & 0xFF, xs & 0xFF, (xe >> 8) & 0xFF, xe & 0xFF);

	write_reg(par, MIPI_DCS_SET_PAGE_ADDRESS,
		  (ys >> 8) & 0xFF, ys & 0xFF, (ye >> 8) & 0xFF, ye & 0xFF);

	write_reg(par, MIPI_DCS_WRITE_MEMORY_START);
}

static void fbtft_reset(struct fbtft_par *par)
{
	if (par->gpio.reset == -1)
		return;
	fbtft_dbg("%s()\n", __func__);
	gpio_set_value(par->gpio.reset, 0);
	udelay(20);
	gpio_set_value(par->gpio.reset, 1);
	mdelay(120);
}

static void fbtft_update_display(struct fbtft_par *par)
{
	size_t len;
	unsigned width, height;

	fbtft_dbg("%s()\n", __func__);

	if ((par->rotate == 90) || (par->rotate == 270)) {
		width = par->display->height;
		height = par->display->width;
	} else {
		width = par->display->width;
		height = par->display->height;
	}

	if (par->display->fbtftops.set_addr_win)
		par->display->fbtftops.set_addr_win(par, 0, 0,
						    width - 1, height - 1);

	len = par->display->height * par->display->width *
		(par->display->bpp / 8);
	if (par->display->fbtftops.write_vmem(par, 0, len) < 0)
		error("write_vmem failed to update display buffer\n");
}

static int fbtft_gpio_out_request(int gpio, const char *label)
{
	if (gpio == -1)
		return 0;

	if (gpio_request(gpio, label)) {
		error("failed to request gpio %s: %u\n", label, gpio);
		return -1;
	}

	return gpio_direction_output(gpio, 1);
}

void lcd_ctrl_init(void *fb)
{
	struct fbtft_par *par = (struct fbtft_par *)panel_info.priv;

	fbtft_dbg("%s(fb=0x%x)\n", __func__, fb);

	par->fb = fb;

	/* Set the the default display operations */
	par->display->fbtftops.write = fbtft_write_spi;
	if ((par->display->regwidth == 8) && (par->display->buswidth == 8)) {
		par->display->fbtftops.write_register = fbtft_write_reg8_bus8;
		par->display->fbtftops.write_vmem = fbtft_write_vmem16_bus8;
	} else {
		error("unsupported fbtft display regwidth/buswidth %d/%d\n",
		      par->display->regwidth, par->display->buswidth);
		return;
	}
	if (!par->display->fbtftops.set_addr_win)
		par->display->fbtftops.set_addr_win = fbtft_set_addr_win;
	if (!par->display->fbtftops.update_display)
		par->display->fbtftops.update_display = fbtft_update_display;
	if (!par->display->fbtftops.reset)
		par->display->fbtftops.reset = fbtft_reset;

	/* Configure the GPIO */
	if (fbtft_gpio_out_request(par->gpio.reset, "fbtft-reset"))
		return;
	if (fbtft_gpio_out_request(par->gpio.dc, "fbtft-dc"))
		return;

	/* Configure the SPI */
	par->slave = spi_setup_slave(par->spi_bus, par->spi_cs,
				     par->spi_max_hz, SPI_MODE_0);

	if (!par->slave) {
		error("PL022 SPI controller init failed\n");
		return;
	}

	if (spi_set_wordlen(par->slave, par->display->buswidth)) {
		error("SPI controller failed to set wordlen\n");
		spi_free_slave(par->slave);
		return;
	}
}

void lcd_enable(void)
{
	struct fbtft_par *par = (struct fbtft_par *)panel_info.priv;

	fbtft_dbg("%s()\n", __func__);

	/* Initialize the display  */
	par->display->fbtftops.init_display(par);

	/* Configure display bgr and rotation */
	if (par->display->fbtftops.set_var)
		par->display->fbtftops.set_var(par);

	/* Configure display gamma */
	if (par->display->fbtftops.set_gamma)
		par->display->fbtftops.set_gamma(par);

	/* Update the display */
	par->display->fbtftops.update_display(par);
}
